import six
import sys
import os

if sys.version_info[0] >= 3.3:
    from types import SimpleNamespace as Dataset
else:
    from argparse import Namespace as Dataset


def associate_by_ext_suffix(datasets):
    has_ext = []
    has_no_ext = []
    for dataset in datasets:
        out_list = has_ext if "_ext" in dataset.name else has_no_ext
        out_list.append(dataset)

    for dataset in has_no_ext:
        associates = [d for d in has_ext if d.name.startswith(dataset.name)]
        associates.append(dataset)
        names = [a.name for a in associates]
        for index in range(len(associates)):
            associates[index].associates = names[:index]
            associates[index].associates += names[index + 1:]


def _load_yaml(path):
    import yaml
    with open(path, 'r') as f:
        datasets_dict = yaml.load(f)
    if not datasets_dict:
        raise RuntimeError("Empty config file in '%s'" % path)
    return datasets_dict


def from_yaml(path, defaults={}, find_associates=associate_by_ext_suffix):
    datasets_dict = _load_yaml(path)
    this_dir = os.path.dirname(path)
    return get_datasets(datasets_dict, defaults, this_dir=this_dir,
                        find_associates=associate_by_ext_suffix)


def get_datasets(datasets_dict, defaults={},
                 find_associates=associate_by_ext_suffix, already_imported=None, this_dir=None):
    datasets = []
    defaults.update(datasets_dict.get("defaults", {}))
    if "import" not in datasets_dict and "datasets" not in datasets_dict:
        raise RuntimeError("Neither 'datasets' nor 'import' were specified in file list")

    if already_imported is None:
        already_imported = set()
    for import_file in datasets_dict.get("import", []):
        if this_dir:
            import_file = import_file.format(this_dir=this_dir)
        if import_file in already_imported:
            continue
        already_imported.add(import_file)
        contents = _load_yaml(import_file)
        datasets += get_datasets(contents, defaults=defaults.copy(), this_dir=os.path.dirname(import_file),
                                 find_associates=find_associates, already_imported=already_imported)
    for dataset in datasets_dict.get("datasets", []):
        if isinstance(dataset, six.string_types):
            dataset_kwargs = _from_string(dataset, defaults)
        elif isinstance(dataset, dict):
            dataset_kwargs = _from_dict(dataset, defaults)
        else:
            raise TypeError("{} not a string or dict".format(dataset))
        datasets.append(Dataset(**dataset_kwargs))

    # Associate samples
    find_associates(datasets)

    return datasets


def _from_string(dataset, default):
    cfg = default.copy()
    cfg["name"] = dataset
    return cfg


def _from_dict(dataset, default):
    cfg = default.copy()
    cfg.update(dataset)
    if "name" not in cfg:
        raise RuntimeError(
            "Dataset provided as dict, without key-value pair for 'name'")
    return cfg



from __future__ import print_function
from . import write
from . import read


def process_args_write(args=None):
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs='*')
    parser.add_argument("-d", "--dataset", required=True,
                        help="Which dataset to associate these files to")
    parser.add_argument("-o", "--output", default="file_list.txt",
                        type=str, help="Name of output file list")
    parser.add_argument("--mc", dest="eventtype", action="store_const", const="mc", default=None,
                        help="Specify if this dataset contains simulated data")
    parser.add_argument("--data", dest="eventtype", action="store_const", const="data",
                        help="Specify if this dataset contains real data")
    parser.add_argument("-t", "--tree-name", default="Events", type=str,
                        help="Provide the name of the tree in the input files to calculate number of events, etc")
    parser.add_argument("-u", "--user", default=[], type=str, action="append",
                        help="Add a user function to extend the dataset dictionary,"
                             " eg. my_package.my_module.some_function")

    def split_meta(arg):
        if "=" not in arg:
            msg = "option not of the form 'key=value'"
            raise argparse.ArgumentTypeError(msg)
        split = arg.split("=", 2)
        return split

    parser.add_argument("-m", "--meta", action="append", type=split_meta, default=[],
                        help="Add other metadata (eg cross-section, run era) for this dataset."
                             + "  Must take the form of 'key=value' ")
    return parser


def main_write(args=None):
    parser = process_args_write(args)
    args = parser.parse_args()

    dataset = write.prepare_file_list(files=args.files, dataset=args.dataset,
                                      eventtype=args.eventtype, tree_name=args.tree_name)
    write.add_meta(dataset, args.meta)
    for user_func in args.user:
        write.process_user_function(dataset, user_func)

    write.write_yaml(dataset, args.output)


def process_args_check(args=None):
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs='*')
    parser.add_argument("-o", "--output", default=None,
                        type=str, help="Name of output file list to expand things to")
    parser.add_argument("-f", "--fields", default="nfiles",
                        type=str, help="Comma-separated list of fields to dump for each dataset ")
    args = parser.parse_args()

    args.fields = args.fields.split(",")
    return args


def main_check(args=None):
    args = process_args_check(args)

    datasets = []
    for infile in args.files:
        datasets += read.from_yaml(infile)

    for dataset in datasets:
        if len(dataset.files) != dataset.nfiles:
            msg = "{d.name}: corrupted dataset: bad nfiles value, should be: {len}, got: {d.nfiles}"
            print(msg.format(d=dataset, len=len(dataset.files)))
            continue
        print("==", dataset.name, "==")
        for field in args.fields:
            msg = "   %(field)s = {d.%(field)s}" % dict(field=field)
            print(msg.format(d=dataset))
    print("Total number of datasets:", len(datasets))

    if args.output:
        datasets = write.prepare_contents(datasets)
        write.write_yaml(datasets, args.output, append=False)



from __future__ import print_function
import itertools
import operator
import os
from fast_curator import read
from collections import defaultdict
import logging
logger = logging.getLogger(__name__)


class UsingUproot():
    import uproot
    from . import xrootd_glob

    @staticmethod
    def expand_file_list(files):
        full_list = []
        for name in files:
            expanded = UsingUproot.xrootd_glob.glob(name)
            full_list += [str(exp) for exp in expanded]
        return full_list

    @staticmethod
    def total_entries(files, tree):
        return UsingUproot.uproot.numentries(files, tree)


class UsingROOT():

    @staticmethod
    def expand_file_list(files):
        from rootpy.utils.ext_glob import glob
        full_list = []
        for name in files:
            expanded = glob(name)
            full_list += [str(exp) for exp in expanded]
        return full_list

    @staticmethod
    def total_entries(files, tree):
        from rootpy import ROOT
        chain = ROOT.TChain(tree)
        for _file in files:
            chain.Add(_file)
        return chain.GetEntries()


def prepare_file_list(files, dataset, eventtype, tree_name, use_uproot=True, absolute_paths=True):
    """
    Expands all globs in the file lists and creates a dataframe similar to those from a DAS query
    """

    process_files = UsingUproot if use_uproot else UsingROOT
    full_list = process_files.expand_file_list(files)
    if absolute_paths:
        full_list = [os.path.realpath(f) if ':' not in f else f for f in full_list]
    numentries = process_files.total_entries(full_list, tree_name)

    data = {}
    data["eventtype"] = eventtype
    data["name"] = dataset
    data["nevents"] = numentries
    data["nfiles"] = len(full_list)
    data["files"] = full_list
    data["tree"] = tree_name

    return data


def select_default(values):
    groups = itertools.groupby(sorted(values))
    groups = [(group, sum(1 for _ in items)) for group, items in groups]
    groups = [group for group in groups if group[1] > 1]
    if not groups:
        return None
    most_common, number_items = max(groups, key=operator.itemgetter(1))
    is_unique = sum([1 for group in groups if group[1] == number_items]) == 1
    if not is_unique:
        return None
    return most_common


def prepare_contents(datasets):
    datasets = [vars(data) if isinstance(data, read.Dataset)
                else data for data in datasets]
    for d in datasets:
        if "associates" in d:
            del d["associates"]

    # build the default properties
    values = defaultdict(list)
    for data in datasets:
        for k, v in data.items():
            values[k].append(v)

    defaults = {}
    for key, vals in values.items():
        if key == "name":
            continue
        is_in_all_datasets = len(vals) == len(datasets)
        if not is_in_all_datasets:
            continue
        new_default = select_default(vals)
        if new_default:
            defaults[key] = new_default

    cleaned_datasets = []
    for data in datasets:
        new_data = {}
        for key, val in data.items():
            if key in defaults and val == defaults[key]:
                continue
            new_data[key] = val
        cleaned_datasets.append(new_data)

    contents = dict(datasets=cleaned_datasets)
    if defaults:
        contents["defaults"] = defaults
    return contents


def write_yaml(dataset, out_file, append=True):
    import yaml
    if os.path.exists(out_file) and append:
        datasets = read.from_yaml(out_file)
        datasets.append(dataset)
        contents = prepare_contents(datasets)
    else:
        contents = {}
        contents["datasets"] = [dataset]

    # https://stackoverflow.com/questions/25108581/python-yaml-dump-bad-indentation
    class MyDumper(yaml.Dumper):
        def increase_indent(self, flow=False, indentless=False):
            return super(MyDumper, self).increase_indent(flow, False)

    yaml_contents = yaml.dump(
        contents, Dumper=MyDumper, default_flow_style=False)
    with open(out_file, 'w') as out:
        out.write(yaml_contents)

    return yaml_contents


def add_meta(dataset, meta):
    for key, value in meta:
        if key in meta:
            msg = "Meta data '%s' will override an existing value" % key
            raise RuntimeError(msg)
        dataset[key] = value


def process_user_function(dataset, user_func):
    import importlib
    path = user_func.split(".")
    mod_name = ".".join(path[:-1])
    module = importlib.import_module(mod_name)

    func_name = path[-1]
    function = getattr(module, func_name)

    function(dataset)



"""
Reproduce the standard glob package behaviour
"""
import glob as gl
import os
import fnmatch
import sys
if sys.version_info[0] > 2:
    from urllib.parse import urlparse
else:
    from urlparse import urlparse


__all__ = ["glob", "iglob"]


def split_url(url):
    parsed_uri = urlparse(url)
    domain = '{uri.scheme}://{uri.netloc}/'.format(uri=parsed_uri)
    path = parsed_uri.path
    return domain, path


def glob(pathname):
    # Let normal python glob try first
    try_glob = gl.glob(pathname)
    if try_glob:
        return try_glob

    # If pathname does not contain a wildcard:
    if not gl.has_magic(pathname):
        return [pathname]

    # Else try xrootd instead
    return xrootd_glob(pathname)


def xrootd_glob(pathname):
    from pyxrootd.client import FileSystem
    # Split the pathname into a directory and basename
    dirs, basename = os.path.split(pathname)

    if gl.has_magic(dirs):
        dirs = xrootd_glob(dirs)
    else:
        dirs = [dirs]

    files = []
    for dirname in dirs:
        host, path = split_url(dirname)
        query = FileSystem(host)

        if not query:
            raise RuntimeError("Cannot prepare xrootd query")

        _, dirlist = query.dirlist(path)
        for entry in dirlist["dirlist"]:
            filename = entry["name"]
            if filename in [".", ".."]:
                continue
            if not fnmatch.fnmatchcase(filename, basename):
                continue
            files.append(os.path.join(dirname, filename))

    return files


def iglob(pathname):
    for name in glob(pathname):
        yield name
